//serialPortSniff.cc
/*************************************************************************
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory, and the Regents of the University of California, as
* Operator of Los Alamos National Laboratory. EPICS BASE Versions 3.13.7
* and higher are distributed subject to a Software License Agreement found
* in file LICENSE that is included with this distribution.
*************************************************************************/

/*
        Author: Marty Kraimer
        Date: 21OCT98
  
        Modifications:
        16-May-2003 Mark Rivers.  Added following features:
            - Output lines now contain the following information in order:
               - The name of the serial port.
               - The time in seconds that the first character on that line was 
                 transmitted or received, in seconds after serialPortSniff was begun.
               - The character 'T' or 'R' indicating if the characters on that line
                 were Transmitted or Received.
               - The characters transmitted or received, up to a maximum of 20 per
                 line.  The characters are first printed in ASCII (if printable) and
                 then printed in hex.
            - A new line is printed whenever any of the following occurs:
               - A change from Transmit to Receive or from Receive to Transmit.
               - 20 characters have been received and are in the buffer.
               - More than 2 seconds of idle time have elapsed since the last time 
                 a line was printed, and there are any characters in the buffer.
            - Increased the stack size from 2000 to 3000, it was marginal on PowerPC.
  
            These changes make serial "conversations" much easier to read.  
            The timestamps are useful in analyzing performance. The idle timeout
            ensures that all characters sent or received are printed within 2 seconds.
        
*/
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>

#include <tickLib.h>
#include <taskLib.h>
#include <rngLib.h>
#include <sysLib.h>


#include "SerialPort.h"

class SniffData {
public:
    bool isReceive;
    unsigned char byte;
    ULONG time;
};

#define RING_BUFF_SIZE 1000
#define MAX_CHARS_PER_LINE 20
#define MAX_IDLE_TIME 2.0
int ringBufSize = RING_BUFF_SIZE * sizeof(SniffData);

class SniffPrivate {
public:
    RING_ID ringbuf;
};

static void peekTask(const char *portName,int nseconds);
static char taskname[] = "serialSniff";

int serialPortSniff(const char *portName,int nseconds);
static void peekHandler(void* parm,bool isReceive, unsigned char byte);
static void printQueue(const char *portName, bool *isReceiveQueue, 
                       unsigned char *byteQueue,
                       double *timeQueue, int len);
static void peekTask(const char *portName,int nseconds);

int serialPortSniff(const char *portName,int nseconds)
{
    taskSpawn(taskname,130,VX_FP_TASK,3000,
        (FUNCPTR)peekTask,
        (int)portName,nseconds,0,0,0,0,0,0,0,0);
    return(0);
}

static void peekHandler(void* parm,bool isReceive, unsigned char byte)
{
    SniffPrivate *pvt = (SniffPrivate *)parm;
    SniffData data;
    data.isReceive = isReceive;
    data.byte = byte;
    data.time = tickGet();
    rngBufPut(pvt->ringbuf,(char *)&data,sizeof(SniffData));
}

static void printQueue(const char *portName, bool *isReceiveQueue, 
                       unsigned char *byteQueue,
                       double *timeQueue, int len)
{
    int i;
    if (len <= 0) return;
    printf("%s %c %.2f ", portName, (isReceiveQueue[0] ? 'R' : 'T'), 
                          timeQueue[0]);
    for(i=0;i<len; i++)
        printf("%c", (isprint(byteQueue[i]) ? byteQueue[i] : '.'));
    for(i=len;i<MAX_CHARS_PER_LINE+1; i++)
        printf("%c", ' ');
    for(i=0;i<len; i++)
        printf("%2.2X ", byteQueue[i] );
    printf("\n");
}

static void peekTask(const char *portName,int nseconds)
{
    SniffPrivate pvt;
    bool isReceiveQueue[MAX_CHARS_PER_LINE];
    unsigned char byteQueue[MAX_CHARS_PER_LINE];
    double timeQueue[MAX_CHARS_PER_LINE];
    int nextQueue = 0;
    ULONG totalTicks = (ULONG)(nseconds*sysClkRateGet());

    if((pvt.ringbuf = rngCreate(sizeof(void *) * ringBufSize))==NULL){
        printf("peekTask: initWatchDog failed\n");
        return;
    }
    if(!SerialPort::setSniffHandler(portName,&pvt,peekHandler)) return;
    ULONG start = ::tickGet();
    bool isReceive=TRUE;
    ULONG maxIdleTicks = (ULONG)(MAX_IDLE_TIME*sysClkRateGet());
    ULONG idleStart=tickGet();
    ULONG idleTicks=0;
    printf("\n");
    while(true) {
        ULONG ticksNow;
        taskDelay(sysClkRateGet());
        idleTicks = tickGet() - idleStart;
        if ((idleTicks > maxIdleTicks) && (nextQueue > 0)) {
            printQueue(portName, isReceiveQueue, byteQueue, 
                       timeQueue, nextQueue);
            nextQueue = 0;
        }

        while (rngNBytes(pvt.ringbuf)>=(int)sizeof(SniffData)){
            SniffData data;
            if (rngBufGet(pvt.ringbuf,(char *)&data,sizeof(SniffData)) !=
                   (int)sizeof(SniffData)) {
                printf("peekTask rngBufGet returned error\n");
            } else {
                if ((isReceive != data.isReceive) || 
                    (nextQueue == MAX_CHARS_PER_LINE)) {
                    printQueue(portName, isReceiveQueue, byteQueue, 
                               timeQueue, nextQueue);
                    nextQueue = 0;
                    idleStart = tickGet();
                }
                isReceive = data.isReceive;
                isReceiveQueue[nextQueue] = data.isReceive;
                byteQueue[nextQueue] = data.byte;
                timeQueue[nextQueue] = (double)(data.time - start) / 
                                                sysClkRateGet();
                nextQueue++;
            }
            ticksNow = tickGet() - start;
            if(ticksNow>=totalTicks) break;
        }
        ticksNow = tickGet() - start;
        if(ticksNow>=totalTicks) break;
    }
    if(nextQueue>0) printQueue(portName, isReceiveQueue, byteQueue,
                               timeQueue, nextQueue);
    printf("serialPortSniff done\n");
    SerialPort::freeSniffHandler(portName);
    rngDelete(pvt.ringbuf);
}
